//
//  AppDelegate.swift
//  DigitalOutput
//
//  Created by Phidgets on 2018-02-15.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = DigitalOutput()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var dutyCycleSlider: NSSlider!
    @IBOutlet weak var minDutyCycleLabel: NSTextField!
    @IBOutlet weak var maxDutyCycleLabel: NSTextField!
    @IBOutlet weak var dutyCycleLabel: NSTextField!
    @IBOutlet weak var forwardVoltagePopUp: NSPopUpButton!
    @IBOutlet weak var currentLimitSlider: NSSlider!
    @IBOutlet weak var minCurrentLimitLabel: NSTextField!
    @IBOutlet weak var maxCurrentLimitLabel: NSTextField!
    @IBOutlet weak var currentLimitLabel: NSTextField!
    @IBOutlet weak var settingsBox: NSBox!
    @IBOutlet weak var extraSettingsBox: NSBox!
    @IBOutlet weak var dutyCycleBox: NSBox!
    @IBOutlet weak var stateButton: NSButton!
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0

    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! DigitalOutput
            let deviceID = try attachedDevice.getDeviceID()
            let subclass = try attachedDevice.getChannelSubclass()
            
            var minDutyCycle:Double?
            var maxDutyCycle:Double?
            var supportedForwardVoltages = [String]()
            var minCurrentLimit:Double?
            var maxCurrentLimit:Double?
            var currentLimit:Double?
            var forwardVoltage:LEDForwardVoltage?
            
            switch(deviceID){
            case DeviceID.PN_1031:
                fallthrough
            case DeviceID.PN_1032:
                supportedForwardVoltages = ["1.7 V", "2.75 V","3.9 V","5.0 V"]
                break
            case DeviceID.PN_LED1000:
                supportedForwardVoltages = ["3.2 V","4.0 V","4.8 V","5.6 V"]
                break
            default:
                break
            }
            
            if(subclass == ChannelSubclass.digitalOutputLEDDriver){
                minCurrentLimit = try attachedDevice.getMinLEDCurrentLimit()
                maxCurrentLimit = try attachedDevice.getMaxLEDCurrentLimit()
                currentLimit = try attachedDevice.getLEDCurrentLimit()
                forwardVoltage = try attachedDevice.getLEDForwardVoltage()
            }
            
            if(subclass == ChannelSubclass.digitalOutputDutyCycle || subclass == ChannelSubclass.digitalOutputLEDDriver){
                minDutyCycle = try attachedDevice.getMinDutyCycle()
                maxDutyCycle = try attachedDevice.getMaxDutyCycle()
            }
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.settingsBox.isHidden = false
                
                if(subclass == ChannelSubclass.digitalOutputDutyCycle || subclass == ChannelSubclass.digitalOutputLEDDriver){
                    self.dutyCycleBox.isHidden = false
                    var frame = self.stateButton.frame
                    frame.origin.x = self.dutyCycleLabel.frame.origin.x + 50
                    self.stateButton.frame = frame
                    self.dutyCycleSlider.minValue = minDutyCycle!
                    self.dutyCycleSlider.maxValue = maxDutyCycle!
                    self.dutyCycleSlider.doubleValue = minDutyCycle!
                    self.minDutyCycleLabel.stringValue = String(format:"%.2f", minDutyCycle!)
                    self.maxDutyCycleLabel.stringValue = String(format:"%.2f", maxDutyCycle!)
                    self.dutyCycleLabel.stringValue = String(format:"%.2f",minDutyCycle!)
                }
                else{
                    self.dutyCycleBox.isHidden = true
                    var frame = self.stateButton.frame
                    frame.origin.x = (self.settingsBox.frame.size.width / 2) - frame.size.width/2
                    self.stateButton.frame = frame
                }
                
                if(subclass == ChannelSubclass.digitalOutputLEDDriver){
                    self.extraSettingsBox.isHidden = false
                    self.forwardVoltagePopUp.addItems(withTitles: supportedForwardVoltages)
                    
                    self.forwardVoltagePopUp.selectItem(withTitle: (forwardVoltage == LEDForwardVoltage.volts_1_7) ? "1.7 V" : (forwardVoltage == LEDForwardVoltage.volts_2_75) ? "2.75 V" : (forwardVoltage == LEDForwardVoltage.volts_3_2) ? "3.2 V" : (forwardVoltage == LEDForwardVoltage.volts_3_9) ? "3.9 V" : (forwardVoltage == LEDForwardVoltage.volts_4_0) ? "4.0 V" : (forwardVoltage == LEDForwardVoltage.volts_4_8) ? "4.8 V" : (forwardVoltage == LEDForwardVoltage.volts_5_0) ? "5.0 V" : "5.6 V")
                    
                    self.currentLimitSlider.minValue = minCurrentLimit!
                    self.currentLimitSlider.maxValue = maxCurrentLimit!
                    self.currentLimitSlider.doubleValue = currentLimit!
                    self.minCurrentLimitLabel.stringValue = String(format:"%.3f",minCurrentLimit!)
                    self.maxCurrentLimitLabel.stringValue = String(format:"%.3f",maxCurrentLimit!)
                    self.currentLimitLabel.stringValue = String(format:"%.3f A", currentLimit!)
                }

                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                if(self.extraSettingsBox.isHidden){
                    frame.size.height -= self.extraSettingsBox.frame.height
                }
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.settingsBox.isHidden = true
            self.extraSettingsBox.isHidden = true
            self.forwardVoltagePopUp.removeAllItems()
            
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    //MARK: GUI Controls
    @IBAction func setDutyCycle(_ sender: Any) {
        let slider = sender as! NSSlider
        do{
            try ch.setDutyCycle(slider.doubleValue)
            dutyCycleLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setState(_ sender: Any) {
        do{
            let state = try ch.getState()
            try ch.setState(!state)
            if(state){
                self.stateButton.title = "On"
                self.dutyCycleSlider.doubleValue = self.dutyCycleSlider.minValue
                self.dutyCycleLabel.stringValue = String(format:"%.2f",self.dutyCycleSlider.minValue)
            }
            else{
                self.stateButton.title = "Off"
                self.dutyCycleSlider.doubleValue = self.dutyCycleSlider.maxValue
                self.dutyCycleLabel.stringValue = String(format:"%.2f",self.dutyCycleSlider.maxValue)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setForwardVoltage(_ sender: Any) {
        let popup = sender as! NSPopUpButton
        do{
            switch(popup.titleOfSelectedItem ?? ""){
            case "1.7 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_1_7)
                break
            case "2.75 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_2_75)
                break
            case "3.2 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_3_2)
                break
            case "3.9 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_3_9)
                break
            case "4.0 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_4_0)
                break
            case "4.8 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_4_8)
                break
            case "5.0 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_5_0)
                break
            case "5.6 V":
                try ch.setLEDForwardVoltage(LEDForwardVoltage.volts_5_6)
                break
            default:
                break
                
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setCurrentLimit(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setLEDCurrentLimit(slider.doubleValue)
            currentLimitLabel.stringValue = String(format:"%.3f A", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }

}

